import java.awt.geom.Point2D;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Properties;
import java.util.Random;

import statistics.Histogram;
//import statistics.LogHistogram;



public class Society implements Serializable{
	private static int RANDOM_TYPE; // how the agents are selected. Default 1 - one according to wealth and the other randomly
	private static int INIT_WEALTH_DISTRIBUTION; // initial wealth distribution 0 - dirac delta distribution, 1 - uniform distribution; 2 - normal distribution

	private static int PAYOFF_FUNCTION; // 0 - 1; 1 - min; 2 - mox; 3 - average; 4 - geomeric average 
	private static double PAYOFF_FACTOR; // multiplies payoffs function

	private static double RANDOM_FACTOR; // randomly changes the wealth.
	
	private static double INCOME_TAX_FACTOR;
	private static double INCOME_TAX_EXPONENT;

	private static double WEALTH_TAX_FACTOR;
	private static double WEALTH_TAX_EXPONENT;
	
	
	private static final long serialVersionUID = 7031507377036548702L;
	
	private static Random rnd;

	static{
		Properties p = new Properties();
		try {
			p.load(new FileInputStream("input.txt"));
			RANDOM_TYPE = Integer.parseInt(p.getProperty("RandomType", "1"));
			INIT_WEALTH_DISTRIBUTION = Integer.parseInt(p.getProperty("InitialWealthDistribution", "0"));

			PAYOFF_FACTOR = Double.parseDouble(p.getProperty("PayoffFactor", "0.01"));
			PAYOFF_FUNCTION = Integer.parseInt(p.getProperty("PayoffFunction", "1"));

			RANDOM_FACTOR = Double.parseDouble(p.getProperty("RandomFactor", "0.0"));
			
			INCOME_TAX_FACTOR = Double.parseDouble(p.getProperty("IncomeTaxFactor", "0.1"));
			INCOME_TAX_EXPONENT = Double.parseDouble(p.getProperty("IncomeTaxExponent", "1.0"));

			WEALTH_TAX_FACTOR = Double.parseDouble(p.getProperty("WealthTaxFactor", "0.0"));
			WEALTH_TAX_EXPONENT = Double.parseDouble(p.getProperty("WealthTaxExponent", "1.0"));
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		rnd = new Random(1l);
	}
	
	
	Agent[] agents;
	private double[] indexmap;
	private ArrayList<HashSet<Integer>> htMaps;
	
	public double max;
	public double gain;
	private double gini, entropy;
	public double[] quantiles = new double[10];
	public int active;
	
	/**
	 * Constuctor
	 * @param n number of agents
	 */
	public Society(int n){
		this.agents = new Agent[n];
		this.indexmap = new double[n];
		this.htMaps = new ArrayList<>(n);
		for(int i=0; i<n; i++){
			this.agents[i] = new Agent();
			this.htMaps.add(new HashSet<Integer>());
			this.htMaps.get(i).clear();
		}
	}
	
	/**
	 * creates index map - accomodated wealth 
	 */
	private void createIndexMap(){
		double total = this.getTotal();
		double s = 0.0;
		int i;
		for(i=0; i<this.agents.length; i++){
			this.indexmap[i]=s;
			s += this.agents[i].wealth/total;
		}
	}
	
	/**
	 * Finds agent index according to a random number s. 
	 * @param rnd random number from [0,1)
	 * @param type type 0 or 1. If type = 0 the agent is selected with uniform probability distribution. Otherwise it is selected according to index map (with probability proportional to its wealth)
	 * @return
	 */
	private int getIndex(int type, HashSet<Integer> reserved){
		double drnd = rnd.nextDouble();
		if (type==0){
			return (int)(this.agents.length * drnd);
		}else{
			if (reserved==null){
				int min = 0; 
				int max = this.indexmap.length-1;
				int mid;
				while((max-min)>1){
					mid = (max+min) / 2;
					if (this.indexmap[mid]<drnd) min = mid;
					else max=mid;
				}
				return min;
			}else{
				double total = this.getTotal();
				double s = 0.0;
				for(int i: reserved)
					total -= this.agents[i].wealth;
				for(int i=0; i<this.agents.length; i++){
					if (!reserved.contains(i)) {
						s += this.agents[i].wealth/total;
						if (s>drnd)
							return i;
					}
				}
			}
		}
		return -1;
	}
	
	/**
	 * Finds randomly two agents and sores their indexes in i12 array.  
	 * @param i12 array for storing indexes.
	 */
	private void getIndexes(int[] i12){
		
		switch(RANDOM_TYPE){
		case 1: // first according to wealth, second uniformly distributed
			i12[0] = this.getIndex(1, null);
			do{ 
				i12[1] = this.getIndex(0, null);
			}while(i12[1]==i12[0]);
			break;
		case 2: // both according to theirs wealth
			i12[0] = this.getIndex(1, null);
			HashSet<Integer> reserved = new HashSet<>();
			if(this.gini>0.95) {
				reserved.add(i12[0]);
				reserved.addAll(this.htMaps.get(i12[0]));
			}
			do{ 
				if (this.gini>0.95)
					i12[1] = this.getIndex(1, reserved);
				else
					i12[1] = this.getIndex(1, null);
			}while(i12[1]==i12[0]);
			break;
		case 3: // both randomly distributed
			i12[0] = this.getIndex(0, null);
			do{ 
				i12[1] = this.getIndex(0, null);
			}while(i12[1]==i12[0]);
			break;
		}
	}
	
	/**
	 * Inits simulation before start
	 */
	public void init(){
		for(int i=0; i<this.agents.length; i++){
			switch(INIT_WEALTH_DISTRIBUTION){
				case 0:
					this.agents[i].wealth = 1.0; //Math.random();
					break;
				case 1:
					this.agents[i].wealth = rnd.nextDouble();
					break;
				case 2:
					this.agents[i].wealth = Math.abs(rnd.nextGaussian());
					break;
				case 3: 
					this.agents[i].wealth = -Math.log(rnd.nextDouble());
					break;
				case 4:
					if (i<0.01*this.agents.length) 
						this.agents[i].wealth=100.0;
					else
						this.agents[i].wealth = 1.0;
					break;
				case 5:
					this.agents[i].wealth = Math.abs(Math.tan(Math.PI*(rnd.nextDouble() - 0.5)));
			}
			this.agents[i].income = 0.0;
		}
		double total = this.getTotal();
		for(int i=0; i<this.agents.length; i++){
			this.agents[i].wealth /= total;
		}
		this.findStatistics();
	}
	
	/**
	 * Returns the total wealth
	 * @return
	 */
	public double getTotal(){
		double s = 0.0;
		for(int i=0; i<this.agents.length; i++){
			s += this.agents[i].wealth;
		}
		return s;		
	}
	
	public double getFractionHaving(double d){
		double[] tmp;
		double s = 0.0;
		int i;
		
		this.normalize();
		tmp = this.createWealthArray();
		Arrays.sort(tmp);
		for(i=1; i<=tmp.length; i++){
			s += tmp[tmp.length-i];
			if (s>=d)
				break;
		}
		return (double)i/(double)tmp.length;
	}
	
	public double getFractionHaving(){
		double[] tmp;
		double s = 0.0, f = 0.0;
		int i;
		
		this.normalize();
		tmp = this.createWealthArray();
		Arrays.sort(tmp);
		for(i=1; i<=tmp.length; i++){
			s += tmp[tmp.length-i];
			f += 1.0/tmp.length;
			if (s+f >= 1.0)
				break;
		}
		return f;
	}

	/**
	 * returns wealth of the i-th individual
	 * @param i
	 * @return
	 */
	public double getWealth(int i){
		return this.agents[i].wealth;
	}

	private double[] createWealthArray(){
		double[] array = new double[this.agents.length];
		for(int i=0; i<this.agents.length; i++){
			array[i] = this.agents[i].wealth;
		}
		return array;
	}
	
	public void findStatistics(){
		int i;
		
		this.gain = 0;
		for(i=0; i<this.agents.length; i++){
			this.gain += this.agents[i].income;
		}
		
		this.normalize();

		double[] tmp = this.createWealthArray();
		Arrays.sort(tmp);
		i=tmp.length-1;
		while(i>0 && tmp[i]>0){
			i--;			
		}
		this.active = i;

		this.max = tmp[tmp.length-1];
		for(i=0; i<quantiles.length-1; i++){
			this.quantiles[i] = tmp[(int)(tmp.length*((double)i/this.quantiles.length))];
		}
		
		double res = 0.0;
		this.normalize();
		for(i=0; i<tmp.length; i++){
			res += i*tmp[i];
		}
		res = 2*res/(tmp.length) - (tmp.length+1)/(tmp.length);
		this.gini = res;
		
		Histogram hist = new Histogram(tmp[0], tmp[tmp.length-1], 100);
		for(i=0; i<tmp.length; i++) {
			hist.add(tmp[i]);
		}
		Point2D[] pa = hist.getAsPoints();
		this.entropy = 0.0;
		for(i=0; i<pa.length; i++){
			if (pa[i].getY()>0) {
				this.entropy -= pa[i].getY()*Math.log(pa[i].getY());
			}
		}
	}
	
	public double getGiniCoefficient(){
		return this.gini;
	}
	
	public double getEntropy() {
		return this.entropy;
	}
		
	public int getRichiestAgentIndex(){
		double max = 0.0;
		int index=0;
		for(int i=0; i<this.agents.length; i++){
			if (this.agents[i].wealth>max){
				max = this.agents[i].wealth;
				index=i;
			}
		}
		return index;		
	}
	
	public int getPoorestAgentIndex(){
		double min = Double.POSITIVE_INFINITY;
		int index=0;
		for(int i=0; i<this.agents.length; i++){
			if (this.agents[i].wealth<min){
				min = this.agents[i].wealth;
				index=i;
			}
		}
		return index;		
	}
	
	public void normalize(){
		double total = this.getTotal();
		for(int i=0; i<this.agents.length; i++){
			this.agents[i].wealth /= total;
		}
	}
		
	private void interact(int i1, int i2){
		double s1 = this.agents[i1].wealth;
		double s2 = this.agents[i2].wealth;
		double transfer = 0.0;
		switch(PAYOFF_FUNCTION){
		case 0:
			transfer = 1.0 / this.agents.length;
			break;
		case 1:
			transfer = Math.min(s1,  s2);
			break;
		case 2:
			transfer = Math.max(s1,  s2);
			break;			
		case 3:
			transfer = (s1+s2)/2.0;
			break;
		case 4:
			transfer = Math.sqrt(s1*s2);
		}
		transfer *= PAYOFF_FACTOR;
		
		this.agents[i1].income += transfer;
		this.agents[i2].income += transfer;
		this.agents[i1].trades++;
		this.agents[i2].trades++;
	}
	
	private void randomize(){
		for(int i=0; i<this.agents.length; i++){
			this.agents[i].wealth *= 1 + ((2*rnd.nextDouble()-1)*RANDOM_FACTOR);
		}		
	}
	
	private void redistributeIncome(){
		double s, sTax = 0.0;
		for(int i=0; i<this.agents.length; i++){
			s = INCOME_TAX_FACTOR*Math.pow(this.agents[i].income, INCOME_TAX_EXPONENT);
			this.agents[i].income -= s;
			sTax += s;
		}
		s = sTax / this.agents.length;
		for(int i=0; i<this.agents.length; i++){
			this.agents[i].income += s; 
		}
	}

	private void redistributeWealth(){
		double s, sTax = 0.0;
		for(int i=0; i<this.agents.length; i++){
			s = WEALTH_TAX_FACTOR*Math.pow(this.agents[i].wealth, WEALTH_TAX_EXPONENT);
			this.agents[i].wealth -= s;
			sTax += s;
		}
		s = sTax / this.agents.length;
		for(int i=0; i<this.agents.length; i++){
			this.agents[i].income += s; 
		}
	}
	
	public void doCycle(){
		int i, i1, i2;
		int[] i12 = new int[2];
		this.createIndexMap();
		for(i=0; i<this.agents.length; i++){
			this.agents[i].income = 0.0;
			this.agents[i].trades = 0;
			this.htMaps.get(i).clear();
		}
		for(i=0; i<this.agents.length/2; i++){
			this.getIndexes(i12);
			i1 = i12[0]; i2 = i12[1];
			if (i1!=i2 && !this.htMaps.get(i1).contains(i2) && !this.htMaps.get(i2).contains(i1) ){
				this.htMaps.get(i1).add(i2);
				this.htMaps.get(i2).add(i1);
				this.interact(i1, i2);
			}else{
				i--;
			}
		}
		if (INCOME_TAX_FACTOR!=0.0)
			this.redistributeIncome();
		if (WEALTH_TAX_FACTOR!=0.0)
			this.redistributeWealth();
		
		for(i=0; i<this.agents.length; i++){
			this.agents[i].wealth += this.agents[i].income;
		}
		if (RANDOM_FACTOR>0.0)
			this.randomize();

		this.findStatistics();
	}
	
	public Histogram toHistogram(boolean normalize){
		double total = this.getTotal();
		Histogram h;
		if (normalize)
			h = new Histogram(0.0, this.agents[this.getRichiestAgentIndex()].wealth/total, this.agents.length/1000);
		else
			h = new Histogram(0.0, this.agents[this.getRichiestAgentIndex()].wealth, this.agents.length/1000);
		for(int i=0; i<this.agents.length; i++){
			if (normalize)
				h.add(this.agents[i].wealth/total);
			else
				h.add(this.agents[i].wealth);
		}
		return h;		
	}
/*	
	public LogHistogram toLogHistogram(){
		double total = this.getTotal();
		LogHistogram h;
		h = new LogHistogram(this.individuals[this.getMin()], this.individuals[this.getMax()], this.individuals.length/2000);
		for(int i=0; i<this.individuals.length; i++){
			h.add(this.individuals[i]);
		}
		return h;		
	}
*/
}
